//在多线程程序中调用fork

#include<pthread.h>
#include<iostream>
#include<unistd.h>
#include<wait.h>

pthread_mutex_t mutex;
//子线程运行的函数。它首先获得互斥锁mutex，然后暂停5s，再释放该互斥锁
void* another(void *arg)
{
    std::cout << "in child thread, lock the mutex\n";
    pthread_mutex_lock( &mutex);
    sleep(5);
    pthread_mutex_unlock(&mutex);
    std::cout << "i unlock this mutex\n";
}

void prepare()
{
    pthread_mutex_lock(&mutex);
}

void infork()
{
    pthread_mutex_unlock(&mutex);
}

int main()
{
    pthread_mutex_init(&mutex, NULL);
    pthread_t id;
    pthread_create(&id, NULL, another, NULL);
    //父进程中的主线程暂停1s,以确保在执行fork操作之前，子线程已经开始运行并获得了互斥量mutex
    sleep(1);
    //专门的函数，以确保fork调用后父进程和子进程都拥有一个清除的锁状态。
    //   函数原型 pthread_atfork(void(*p)(void), void(*parent)(void), void(*child)(void));
    pthread_atfork(prepare, infork, infork); 

    int pid = fork();
    if(pid < 0)
    {
        pthread_join(id, NULL);
        pthread_mutex_destroy(&mutex);
        return 1;
    }
    else if(pid == 0)
    {
        std::cout << "i am in the child, want to get the lock\n";
        /*
        子线程从父进程继承了互斥锁mutex的状态，该互斥锁处于锁住的状态，
        这是由父进程中的子线程执行pthread_mutex_lock引起的，因此，下面这句加锁操作会一直阻塞
        尽管从逻辑上来说它是不应该阻塞的
        */
        //因为子进程继承了父进程的锁（该锁是锁住的状态，子进程也继承了锁住的状态），这个时候再加锁会造成死锁
        pthread_mutex_lock(&mutex);
        std::cout << "i can not run to here, oop...\n";
        pthread_mutex_unlock(&mutex);
        exit(0);
    }
    else
        wait(NULL);
    pthread_join(id, NULL);
    pthread_mutex_destroy(&mutex);
    return 0;
}